@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
+# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -13,7 +13,7 @@
 ##############################################################################
 """Database connection support
 
-$Id: Connection.py 116685 2010-09-20 19:58:45Z jim $"""
+$Id: Connection.py 111314 2010-04-23 17:30:12Z jim $"""
 
 import logging
 import sys
@@ -22,7 +22,7 @@ import threading
 import warnings
 import os
 import shutil
-from time import time
+import time
 
 from persistent import PickleCache
 
@@ -30,7 +30,8 @@ from persistent import PickleCache
 from persistent.interfaces import IPersistentDataManager
 from ZODB.interfaces import IConnection
 from ZODB.interfaces import IBlobStorage
-from ZODB.blob import Blob, rename_or_copy_blob
+from ZODB.interfaces import IMVCCStorage
+from ZODB.blob import Blob, rename_or_copy_blob, remove_committed_dir
 from transaction.interfaces import ISavepointDataManager
 from transaction.interfaces import IDataManagerSavepoint
 from transaction.interfaces import ISynchronizer
@@ -38,13 +39,14 @@ from zope.interface import implements
 
 import transaction
 
+import ZODB
 from ZODB.blob import SAVEPOINT_SUFFIX
 from ZODB.ConflictResolution import ResolvedSerial
 from ZODB.ExportImport import ExportImport
 from ZODB import POSException
 from ZODB.POSException import InvalidObjectReference, ConnectionStateError
 from ZODB.POSException import ConflictError, ReadConflictError
-from ZODB.POSException import Unsupported
+from ZODB.POSException import Unsupported, ReadOnlyHistoryError
 from ZODB.POSException import POSKeyError
 from ZODB.serialize import ObjectWriter, ObjectReader, myhasattr
 from ZODB.utils import p64, u64, z64, oid_repr, positive_id
@@ -79,25 +81,36 @@ class Connection(ExportImport, object):
     ##########################################################################
     # Connection methods, ZODB.IConnection
 
-    def __init__(self, db, version='', cache_size=400):
+    def __init__(self, db, cache_size=400, before=None, cache_size_bytes=0):
         """Create a new Connection."""
 
         self._log = logging.getLogger('ZODB.Connection')
         self._debug_info = ()
 
         self._db = db
+
+        # historical connection
+        self.before = before
+
         # Multi-database support
         self.connections = {self._db.database_name: self}
 
-        self._version = version
-        self._normal_storage = self._storage = db._storage
-        self.new_oid = db._storage.new_oid
+        storage = db.storage
+        if IMVCCStorage.providedBy(storage):
+            # Use a connection-specific storage instance.
+            self._mvcc_storage = True
+            storage = storage.new_instance()
+        else:
+            self._mvcc_storage = False
+
+        self._normal_storage = self._storage = storage
+        self.new_oid = storage.new_oid
         self._savepoint_storage = None
 
         # Do we need to join a txn manager?
         self._needs_to_join = True
         self.transaction_manager = None
-        self._opened = None # time.time() when DB.open() opened us
+        self.opened = None # time.time() when DB.open() opened us
 
         self._reset_counter = global_reset_counter
         self._load_count = 0   # Number of objects unghosted
@@ -106,19 +119,12 @@ class Connection(ExportImport, object):
         # Cache which can ghostify (forget the state of) objects not
         # recently used. Its API is roughly that of a dict, with
         # additional gc-related and invalidation-related methods.
-        self._cache = PickleCache(self, cache_size)
+        self._cache = PickleCache(self, cache_size, cache_size_bytes)
 
         # The pre-cache is used by get to avoid infinite loops when
         # objects immediately load their state whern they get their
         # persistent data set.
         self._pre_cache = {}
-        
-        if version:
-            # Caches for versions end up empty if the version
-            # is not used for a while. Non-version caches
-            # keep their content indefinitely.
-            # Unclear:  Why do we want version caches to behave this way?
-            self._cache.cache_drain_resistance = 100
 
         # List of all objects (not oids) registered as modified by the
         # persistence machinery, or by add(), or whose access caused a
@@ -140,7 +146,7 @@ class Connection(ExportImport, object):
         # During commit, all objects go to either _modified or _creating:
 
         # Dict of oid->flag of new objects (without serial), either
-        # added by add() or implicitely added (discovered by the
+        # added by add() or implicitly added (discovered by the
         # serializer during commit). The flag is True for implicit
         # adding. Used during abort to remove created objects from the
         # _cache, and by persistent_id to check that a new object isn't
@@ -151,7 +157,6 @@ class Connection(ExportImport, object):
         # in the cache on abort and in other connections on finish.
         self._modified = []
 
-
         # _invalidated queues invalidate messages delivered from the DB
         # _inv_lock prevents one thread from modifying the set while
         # another is processing invalidations.  All the invalidations
@@ -182,12 +187,10 @@ class Connection(ExportImport, object):
         # _conflicts).
         self._conflicts = {}
 
-        # If MVCC is enabled, then _mvcc is True and _txn_time stores
-        # the upper bound on transactions visible to this connection.
-        # That is, all object revisions must be written before _txn_time.
-        # If it is None, then the current revisions are acceptable.
-        # If the connection is in a version, mvcc will be disabled, because
-        # loadBefore() only returns non-version data.
+        # _txn_time stores the upper bound on transactions visible to
+        # this connection. That is, all object revisions must be
+        # written before _txn_time. If it is None, then the current
+        # revisions are acceptable.
         self._txn_time = None
 
         # To support importFile(), implemented in the ExportImport base
@@ -201,7 +204,7 @@ class Connection(ExportImport, object):
 
     def add(self, obj):
         """Add a new object 'obj' to the database and assign it an oid."""
-        if self._opened is None:
+        if self.opened is None:
             raise ConnectionStateError("The database connection is closed")
 
         marker = object()
@@ -225,7 +228,7 @@ class Connection(ExportImport, object):
 
     def get(self, oid):
         """Return the persistent object with oid 'oid'."""
-        if self._opened is None:
+        if self.opened is None:
             raise ConnectionStateError("The database connection is closed")
 
         obj = self._cache.get(oid, None)
@@ -233,15 +236,18 @@ class Connection(ExportImport, object):
             return obj
         obj = self._added.get(oid, None)
         if obj is not None:
-            return obj        
+            return obj
         obj = self._pre_cache.get(oid, None)
         if obj is not None:
             return obj
 
         # This appears to be an MVCC violation because we are loading
         # the must recent data when perhaps we shouldnt. The key is
-        # that we are only creating a ghost!        
-        p, serial = self._storage.load(oid, self._version)
+        # that we are only creating a ghost!
+        # A disadvantage to this optimization is that _p_serial cannot be
+        # trusted until the object has been loaded, which affects both MVCC
+        # and historical connections.
+        p, serial = self._storage.load(oid, '')
         obj = self._reader.getGhost(p)
 
         # Avoid infiniate loop if obj tries to load its state before
@@ -294,45 +300,48 @@ class Connection(ExportImport, object):
 
         self._debug_info = ()
 
-        if self._opened:
+        if self.opened:
             self.transaction_manager.unregisterSynch(self)
 
+        if self._mvcc_storage:
+            self._storage.sync(force=False)
+
         if primary:
             for connection in self.connections.values():
                 if connection is not self:
                     connection.close(False)
 
             # Return the connection to the pool.
-            if self._opened is not None:
+            if self.opened is not None:
                 self._db._returnToPool(self)
 
-                # _returnToPool() set self._opened to None.
+                # _returnToPool() set self.opened to None.
                 # However, we can't assert that here, because self may
                 # have been reused (by another thread) by the time we
                 # get back here.
         else:
-            self._opened = None
-            am = self._db._activity_monitor
-            if am is not None:
-                am.closedConnection(self)
+            self.opened = None
 
     def db(self):
         """Returns a handle to the database this connection belongs to."""
         return self._db
 
     def isReadOnly(self):
-        """Returns True if the storage for this connection is read only."""
-        if self._opened is None:
+        """Returns True if this connection is read only."""
+        if self.opened is None:
             raise ConnectionStateError("The database connection is closed")
-        return self._storage.isReadOnly()
+        return self.before is not None or self._storage.isReadOnly()
 
     def invalidate(self, tid, oids):
         """Notify the Connection that transaction 'tid' invalidated oids."""
+        if self.before is not None:
+            # this is an historical connection.  Invalidations are irrelevant.
+            return
         self._inv_lock.acquire()
         try:
             if self._txn_time is None:
                 self._txn_time = tid
-            elif (tid < self._txn_time) and (tid is not None):
+            elif tid < self._txn_time:
                 raise AssertionError("invalidations out of order, %r < %r"
                                      % (tid, self._txn_time))
 
@@ -347,16 +356,10 @@ class Connection(ExportImport, object):
         finally:
             self._inv_lock.release()
 
-
+    @property
     def root(self):
         """Return the database root object."""
-        return self.get(z64)
-
-    def getVersion(self):
-        """Returns the version this connection is attached to."""
-        if self._storage is None:
-            raise ConnectionStateError("The database connection is closed")
-        return self._version
+        return RootConvenience(self.get(z64))
 
     def get_connection(self, database_name):
         """Return a Connection for the named database."""
@@ -364,7 +367,7 @@ class Connection(ExportImport, object):
         if connection is None:
             new_con = self._db.databases[database_name].open(
                 transaction_manager=self.transaction_manager,
-                version=self._version,
+                before=self.before,
                 )
             self.connections.update(new_con.connections)
             new_con.connections = self.connections
@@ -475,6 +478,16 @@ class Connection(ExportImport, object):
 
     # Process pending invalidations.
     def _flush_invalidations(self):
+        if self._mvcc_storage:
+            # Poll the storage for invalidations.
+            invalidated = self._storage.poll_invalidations()
+            if invalidated is None:
+                # special value: the transaction is so old that
+                # we need to flush the whole cache.
+                self._cache.invalidate(self._cache.cache_data.keys())
+            elif invalidated:
+                self._cache.invalidate(invalidated)
+
         self._inv_lock.acquire()
         try:
             # Non-ghostifiable objects may need to read when they are
@@ -547,6 +560,9 @@ class Connection(ExportImport, object):
     def _commit(self, transaction):
         """Commit changes to an object"""
 
+        if self.before is not None:
+            raise ReadOnlyHistoryError()
+
         if self._import:
             # We are importing an export file. We alsways do this
             # while making a savepoint so we can copy export data
@@ -563,7 +579,7 @@ class Connection(ExportImport, object):
         self._added_during_commit = []
 
         if self._invalidatedCache:
-            raise ConflictError()            
+            raise ConflictError()
 
         for obj in self._registered_objects:
             oid = obj._p_oid
@@ -605,7 +621,7 @@ class Connection(ExportImport, object):
                  or self._savepoint_storage.creating[oid]
                  )
                 ):
-                
+
                 # obj is a new object
 
                 # Because obj was added, it is now in _creating, so it
@@ -626,21 +642,21 @@ class Connection(ExportImport, object):
             if isinstance(obj, Blob):
                 if not IBlobStorage.providedBy(self._storage):
                     raise Unsupported(
-                        "Storing Blobs in %s is not supported." % 
+                        "Storing Blobs in %s is not supported." %
                         repr(self._storage))
                 if obj.opened():
                     raise ValueError("Can't commit with opened blobs.")
                 s = self._storage.storeBlob(oid, serial, p,
                                             obj._uncommitted(),
-                                            self._version, transaction)
+                                            '', transaction)
                 # we invalidate the object here in order to ensure
                 # that that the next attribute access of its name
                 # unghostify it, which will cause its blob data
                 # to be reattached "cleanly"
                 obj._p_invalidate()
             else:
-                s = self._storage.store(oid, serial, p, self._version,
-                                        transaction)
+                s = self._storage.store(oid, serial, p, '', transaction)
+
             self._store_count += 1
             # Put the object in the cache before handling the
             # response, just in case the response contains the
@@ -655,6 +671,9 @@ class Connection(ExportImport, object):
                 else:
                     raise
 
+            self._cache.update_object_size_estimation(oid, len(p))
+            obj._p_estimated_size = len(p)
+
             self._handle_serial(s, oid)
 
     def _handle_serial(self, store_return, oid=None, change=1):
@@ -752,6 +771,10 @@ class Connection(ExportImport, object):
         """Indicate confirmation that the transaction is done."""
 
         def callback(tid):
+            if self._mvcc_storage:
+                # Inter-connection invalidation is not needed when the
+                # storage provides MVCC.
+                return
             d = dict.fromkeys(self._modified)
             self._db.invalidate(tid, d, self)
 #       It's important that the storage calls the passed function
@@ -805,7 +828,7 @@ class Connection(ExportImport, object):
         the database."""
         oid = obj._p_oid
 
-        if self._opened is None:
+        if self.opened is None:
             msg = ("Shouldn't load state for %s "
                    "when the connection is closed" % oid_repr(oid))
             self._log.error(msg)
@@ -836,43 +859,54 @@ class Connection(ExportImport, object):
         # 3. Raise ConflictError.
 
         # Does anything actually use _p_independent()?  It would simplify
-        # the code if we could drop support for it.  
+        # the code if we could drop support for it.
         # (BTrees.Length does.)
 
-        # There is a harmless data race with self._invalidated.  A
-        # dict update could go on in another thread, but we don't care
-        # because we have to check again after the load anyway.
 
+        if self.before is not None:
+            # Load data that was current before the time we have.
+            before = self.before
+            t = self._storage.loadBefore(obj._p_oid, before)
+            if t is None:
+                raise POSKeyError() # historical connection!
+            p, serial, end = t
 
-        if self._invalidatedCache:
-            raise ReadConflictError()
+        else:
+            # There is a harmless data race with self._invalidated.  A
+            # dict update could go on in another thread, but we don't care
+            # because we have to check again after the load anyway.
 
-        if (obj._p_oid in self._invalidated and
-                not myhasattr(obj, "_p_independent")):
-            # If the object has _p_independent(), we will handle it below.
-            self._load_before_or_conflict(obj)
-            return
+            if self._invalidatedCache:
+                raise ReadConflictError()
 
-        p, serial = self._storage.load(obj._p_oid, self._version)
-        self._load_count += 1
-
-        self._inv_lock.acquire()
-        try:
-            invalid = obj._p_oid in self._invalidated
-        finally:
-            self._inv_lock.release()
-
-        if invalid:
-            if myhasattr(obj, "_p_independent"):
-                # This call will raise a ReadConflictError if something
-                # goes wrong
-                self._handle_independent(obj)
-            else:
+            if (obj._p_oid in self._invalidated and
+                    not myhasattr(obj, "_p_independent")):
+                # If the object has _p_independent(), we will handle it below.
                 self._load_before_or_conflict(obj)
                 return
 
+            p, serial = self._storage.load(obj._p_oid, '')
+            self._load_count += 1
+
+            self._inv_lock.acquire()
+            try:
+                invalid = obj._p_oid in self._invalidated
+            finally:
+                self._inv_lock.release()
+
+            if invalid:
+                if myhasattr(obj, "_p_independent"):
+                    # This call will raise a ReadConflictError if something
+                    # goes wrong
+                    self._handle_independent(obj)
+                else:
+                    self._load_before_or_conflict(obj)
+                    return
+
         self._reader.setGhostState(obj, p)
         obj._p_serial = serial
+        self._cache.update_object_size_estimation(obj._p_oid, len(p))
+        obj._p_estimated_size = len(p)
 
         # Blob support
         if isinstance(obj, Blob):
@@ -881,7 +915,7 @@ class Connection(ExportImport, object):
 
     def _load_before_or_conflict(self, obj):
         """Load non-current state for obj or raise ReadConflictError."""
-        if not ((not self._version) and self._setstate_noncurrent(obj)):
+        if not self._setstate_noncurrent(obj):
             self._register(obj)
             self._conflicts[obj._p_oid] = True
             raise ReadConflictError(object=obj)
@@ -908,6 +942,12 @@ class Connection(ExportImport, object):
         assert self._txn_time <= end, (u64(self._txn_time), u64(end))
         self._reader.setGhostState(obj, data)
         obj._p_serial = start
+
+        # MVCC Blob support
+        if isinstance(obj, Blob):
+            obj._p_blob_uncommitted = None
+            obj._p_blob_committed = self._storage.loadBlob(obj._p_oid, start)
+
         return True
 
     def _handle_independent(self, obj):
@@ -998,7 +1038,7 @@ class Connection(ExportImport, object):
         register for afterCompletion() calls.
         """
 
-        self._opened = time()
+        self.opened = time.time()
 
         if transaction_manager is None:
             transaction_manager = transaction.manager
@@ -1031,17 +1071,21 @@ class Connection(ExportImport, object):
         self._invalidated.clear()
         self._invalidatedCache = False
         cache_size = self._cache.cache_size
-        self._reader._cache = self._cache = PickleCache(self, cache_size)
+        cache_size_bytes = self._cache.cache_size_bytes
+        self._cache = cache = PickleCache(self, cache_size, cache_size_bytes)
+        if getattr(self, '_reader', None) is not None:
+            self._reader._cache = cache
+
+    def _releaseStorage(self):
+        """Tell the storage to release resources it's using"""
+        if self._mvcc_storage:
+            self._storage.release()
 
     ##########################################################################
     # Python protocol
 
     def __repr__(self):
-        if self._version:
-            ver = ' (in version %s)' % `self._version`
-        else:
-            ver = ''
-        return '<Connection at %08x%s>' % (positive_id(self), ver)
+        return '<Connection at %08x>' % (positive_id(self),)
 
     # Python protocol
     ##########################################################################
@@ -1051,17 +1095,6 @@ class Connection(ExportImport, object):
 
     __getitem__ = get
 
-    def modifiedInVersion(self, oid):
-        """Returns the version the object with the given oid was modified in.
-
-        If it wasn't modified in a version, the current version of this
-        connection is returned.
-        """
-        try:
-            return self._db.modifiedInVersion(oid)
-        except KeyError:
-            return self.getVersion()
-
     def exchange(self, old, new):
         # called by a ZClasses method that isn't executed by the test suite
         oid = old._p_oid
@@ -1087,7 +1120,7 @@ class Connection(ExportImport, object):
 
     def savepoint(self):
         if self._savepoint_storage is None:
-            tmpstore = TmpStore(self._version, self._normal_storage)
+            tmpstore = TmpStore(self._normal_storage)
             self._savepoint_storage = tmpstore
             self._storage = self._savepoint_storage
 
@@ -1110,8 +1143,9 @@ class Connection(ExportImport, object):
         self._abort()
         self._registered_objects = []
         src = self._storage
-        self._cache.invalidate(src.index)
+        index = src.index
         src.reset(*state)
+        self._cache.invalidate(index)
 
     def _commit_savepoint(self, transaction):
         """Commit all changes made in savepoints and begin 2-phase commit
@@ -1129,18 +1163,22 @@ class Connection(ExportImport, object):
 
         for oid in oids:
             data, serial = src.load(oid, src)
+            obj = self._cache.get(oid, None)
+            if obj is not None:
+                self._cache.update_object_size_estimation(obj._p_oid, len(data))
+                obj._p_estimated_size = len(data)
             if isinstance(self._reader.getGhost(data), Blob):
                 blobfilename = src.loadBlob(oid, serial)
                 s = self._storage.storeBlob(oid, serial, data, blobfilename,
-                                            self._version, transaction)
+                                            '', transaction)
                 # we invalidate the object here in order to ensure
                 # that that the next attribute access of its name
                 # unghostify it, which will cause its blob data
                 # to be reattached "cleanly"
-                self.invalidate(None, (oid, ))
+                self.invalidate(s, {oid:True})
             else:
                 s = self._storage.store(oid, serial, data,
-                                        self._version, transaction)
+                                        '', transaction)
 
             self._handle_serial(s, oid, change=False)
         src.close()
@@ -1190,23 +1228,14 @@ class TmpStore:
 
     implements(IBlobStorage)
 
-    def __init__(self, base_version, storage):
+    def __init__(self, storage):
         self._storage = storage
         for method in (
             'getName', 'new_oid', 'getSize', 'sortKey', 'loadBefore',
+            'isReadOnly'
             ):
             setattr(self, method, getattr(storage, method))
 
-        try:
-            supportsVersions = storage.supportsVersions
-        except AttributeError:
-            pass
-        else:
-            if supportsVersions():
-                self.modifiedInVersion = storage.modifiedInVersion
-                self.versionEmpty = storage.versionEmpty
-
-        self._base_version = base_version
         self._file = tempfile.TemporaryFile()
         # position: current file position
         # _tpos: file position at last commit point
@@ -1215,16 +1244,21 @@ class TmpStore:
         self.index = {}
         self.creating = {}
 
+        self._blob_dir = None
+
     def __len__(self):
         return len(self.index)
 
     def close(self):
         self._file.close()
+        if self._blob_dir is not None:
+            remove_committed_dir(self._blob_dir)
+            self._blob_dir = None
 
     def load(self, oid, version):
         pos = self.index.get(oid)
         if pos is None:
-            return self._storage.load(oid, self._base_version)
+            return self._storage.load(oid, '')
         self._file.seek(pos)
         h = self._file.read(8)
         oidlen = u64(h)
@@ -1239,7 +1273,7 @@ class TmpStore:
     def store(self, oid, serial, data, version, transaction):
         # we have this funny signature so we can reuse the normal non-commit
         # commit logic
-        assert version == self._base_version
+        assert version == ''
         self._file.seek(self.position)
         l = len(data)
         if serial is None:
@@ -1253,7 +1287,8 @@ class TmpStore:
 
     def storeBlob(self, oid, serial, data, blobfilename, version,
                   transaction):
-        serial = self.store(oid, serial, data, version, transaction)
+        assert version == ''
+        serial = self.store(oid, serial, data, '', transaction)
 
         targetpath = self._getBlobPath()
         if not os.path.exists(targetpath):
@@ -1274,13 +1309,27 @@ class TmpStore:
             return self._storage.loadBlob(oid, serial)
         return filename
 
+    def openCommittedBlobFile(self, oid, serial, blob=None):
+        blob_filename = self.loadBlob(oid, serial)
+        if blob is None:
+            return open(blob_filename, 'rb')
+        else:
+            return ZODB.blob.BlobFile(blob_filename, 'r', blob)
+
     def _getBlobPath(self):
-        return os.path.join(self.temporaryDirectory(), 'savepoints')
+        blob_dir = self._blob_dir
+        if blob_dir is None:
+            blob_dir = tempfile.mkdtemp(dir=self.temporaryDirectory(),
+                                        prefix='savepoints')
+            self._blob_dir = blob_dir
+        return blob_dir
 
     def _getCleanFilename(self, oid, tid):
-        return os.path.join(self._getBlobPath(),
-                            "%s-%s%s" % (utils.oid_repr(oid), utils.tid_repr(tid), SAVEPOINT_SUFFIX,)
-                            )
+        return os.path.join(
+            self._getBlobPath(),
+            "%s-%s%s" % (utils.oid_repr(oid), utils.tid_repr(tid),
+                         SAVEPOINT_SUFFIX,)
+            )
 
     def temporaryDirectory(self):
         return self._storage.temporaryDirectory()
@@ -1298,3 +1347,32 @@ class TmpStore:
         # a copy of the index here.  An alternative would be to ensure that
         # all callers pass copies.  As is, our callers do not make copies.
         self.index = index.copy()
+
+class RootConvenience(object):
+
+    def __init__(self, root):
+        self.__dict__['_root'] = root
+
+    def __getattr__(self, name):
+        try:
+            return self._root[name]
+        except KeyError:
+            raise AttributeError(name)
+
+    def __setattr__(self, name, v):
+        self._root[name] = v
+
+    def __delattr__(self, name):
+        try:
+            del self._root[name]
+        except KeyError:
+            raise AttributeError(name)
+
+    def __call__(self):
+        return self._root
+
+    def __repr__(self):
+        names = " ".join(sorted(self._root))
+        if len(names) > 60:
+            names = names[:57].rsplit(' ', 1)[0] + ' ...'
+        return "<root: %s>" % names
