@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2004 Zope Foundation and Contributors.
+# Copyright (c) 2004 Zope Corporation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -13,11 +13,12 @@
 ##############################################################################
 """Tests of savepoint feature
 
-$Id: testConnectionSavepoint.py 113732 2010-06-21 15:31:11Z ctheune $
+$Id: testConnectionSavepoint.py 111314 2010-04-23 17:30:12Z jim $
 """
 import unittest
 from zope.testing import doctest
-import persistent.dict, transaction
+import persistent.mapping
+import transaction
 
 def testAddingThenModifyThenAbort():
     """\
@@ -38,7 +39,7 @@ savepoint.
     >>> connection = db.open()
     >>> root = connection.root()
 
-    >>> ob = persistent.dict.PersistentDict()
+    >>> ob = persistent.mapping.PersistentMapping()
     >>> root['ob'] = ob
     >>> sp = transaction.savepoint()
     >>> ob.x = 1
@@ -138,10 +139,56 @@ Verify all the values are as expected:
     >>> db.close()
 """
 
+def testIsReadonly():
+    """\
+The connection isReadonly method relies on the _storage to have an isReadOnly.
+We simply rely on the underlying storage method.
+
+    >>> import ZODB.tests.util
+    >>> db = ZODB.tests.util.DB()
+    >>> connection = db.open()
+    >>> root = connection.root()
+    >>> root['a'] = 1
+    >>> sp = transaction.savepoint()
+    >>> connection.isReadOnly()
+    False
+"""
+
+class SelfActivatingObject(persistent.Persistent):
+
+    def _p_invalidate(self):
+        super(SelfActivatingObject, self)._p_invalidate()
+        self._p_activate()
+
+def testInvalidateAfterRollback():
+    """\
+The rollback used to invalidate objects before resetting the TmpStore.
+This caused problems for custom _p_invalidate methods that would load
+the wrong state.
+
+    >>> import ZODB.tests.util
+    >>> db = ZODB.tests.util.DB()
+    >>> connection = db.open()
+    >>> root = connection.root()
+
+    >>> root['p'] = p = SelfActivatingObject()
+    >>> transaction.commit()
+    >>> p.foo = 1
+    >>> sp = transaction.savepoint()
+    >>> p.foo = 2
+    >>> sp2 = transaction.savepoint()
+    >>> sp.rollback()
+    >>> p.foo  # This used to wrongly return 2
+    1
+    """
+
+def tearDown(test):
+    transaction.abort()
+
 def test_suite():
     return unittest.TestSuite((
-        doctest.DocFileSuite('testConnectionSavepoint.txt'),
-        doctest.DocTestSuite(),
+        doctest.DocFileSuite('testConnectionSavepoint.txt', tearDown=tearDown),
+        doctest.DocTestSuite(tearDown=tearDown),
         ))
 
 if __name__ == '__main__':
