@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2003 Zope Foundation and Contributors.
+# Copyright (c) 2003 Zope Corporation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -13,12 +13,13 @@
 ##############################################################################
 """ZODB-based persistent weakrefs
 
-$Id: wref.py 113732 2010-06-21 15:31:11Z ctheune $
+$Id: wref.py 111320 2010-04-23 18:12:57Z jim $
 """
 
 __docformat__ = "reStructuredText"
 
 from persistent import Persistent
+import transaction
 
 WeakRefMarker = object()
 
@@ -60,7 +61,7 @@ class WeakRef(object):
     >>> conn1 = db.open()
     >>> conn1.root()['ob'] = ob
     >>> conn1.root()['ref'] = ref
-    >>> ZODB.tests.util.commit()
+    >>> transaction.commit()
 
     If we open a new connection, we can use the reference:
 
@@ -73,7 +74,7 @@ class WeakRef(object):
     But if we delete the referenced object and pack:
 
     >>> del conn2.root()['ob']
-    >>> ZODB.tests.util.commit()
+    >>> transaction.commit()
     >>> ZODB.tests.util.pack(db)
 
     And then look in a new connection:
@@ -98,6 +99,61 @@ class WeakRef(object):
     Always explicitly close databases: :)
 
     >>> db.close()
+    >>> del ob, ref, db, conn1, conn2, conn3
+
+    When multiple databases are in use, a weakref in one database may
+    point to an object in a different database.  Let's create two new
+    databases to demonstrate this.
+
+    >>> dbA = ZODB.tests.util.DB(
+    ...     database_name = 'dbA',
+    ...     )
+    >>> dbB = ZODB.tests.util.DB(
+    ...     database_name = 'dbB',
+    ...     databases = dbA.databases,
+    ...     )
+    >>> connA1 = dbA.open()
+    >>> connB1 = connA1.get_connection('dbB')
+
+    Now create and add a new object and a weak reference, and add them
+    to different databases.
+
+    >>> ob = ZODB.tests.MinPO.MinPO()
+    >>> ref = WeakRef(ob)
+    >>> connA1.root()['ob'] = ob
+    >>> connA1.add(ob)
+    >>> connB1.root()['ref'] = ref
+    >>> transaction.commit()
+
+    After a succesful commit, the reference should know the oid,
+    database name and connection of the object.
+
+    >>> ref.oid == ob._p_oid
+    True
+    >>> ref.database_name == 'dbA'
+    True
+    >>> ref.dm is ob._p_jar is connA1
+    True
+
+    If we open new connections, we should be able to use the reference.
+
+    >>> connA2 = dbA.open()
+    >>> connB2 = connA2.get_connection('dbB')
+    >>> ref2 = connB2.root()['ref']
+    >>> ob2 = connA2.root()['ob']
+    >>> ref2() is ob2
+    True
+    >>> ref2.oid == ob2._p_oid
+    True
+    >>> ref2.database_name == 'dbA'
+    True
+    >>> ref2.dm is ob2._p_jar is connA2
+    True
+
+    Always explicitly close databases: :)
+
+    >>> dbA.close()
+    >>> dbB.close()
 
     """
 
@@ -109,6 +165,8 @@ class WeakRef(object):
         self._v_ob = ob
         self.oid = ob._p_oid
         self.dm = ob._p_jar
+        if self.dm is not None:
+            self.database_name = self.dm.db().database_name
 
     def __call__(self):
         try:
@@ -116,7 +174,7 @@ class WeakRef(object):
         except AttributeError:
             try:
                 self._v_ob = self.dm[self.oid]
-            except KeyError:
+            except (KeyError, AttributeError):
                 return None
             return self._v_ob
 
@@ -180,7 +238,7 @@ class PersistentWeakKeyDictionary(Persistent):
     >>> conn1.root()['d'] = d
     >>> conn1.root()['p2'] = p2
     >>> conn1.root()['p3'] = p3
-    >>> ZODB.tests.util.commit()
+    >>> transaction.commit()
 
     And things still work, as before:
 
@@ -210,7 +268,7 @@ class PersistentWeakKeyDictionary(Persistent):
     from the dictionary:
 
     >>> del conn2.root()['p2']
-    >>> ZODB.tests.util.commit()
+    >>> transaction.commit()
 
     And pack the database, so that the no-longer referenced p2 is
     actually removed from the database.
